import json
from csv import reader as csv_reader, writer as csv_writer
from io import StringIO
from typing import TextIO, BinaryIO, Union, List, Optional, Any, Dict, Iterable

from pynmrstar import definitions, entry as entry_mod, loop as loop_mod, parser as parser_mod, utils
from pynmrstar._internal import _get_comments, _json_serialize, _interpret_file
from pynmrstar.exceptions import FormattingError, IllegalActionError
from pynmrstar.schema import Schema


class Saveframe(object):
    """A saveframe object. Create using the class methods, see below."""

    def __contains__(self, item: str) -> bool:
        """ Check if the saveframe contains a tag or a loop name."""

        try:
            self.__getitem__(item)
            return True
        except KeyError:
            return False

    def __delitem__(self, item: Union[int, str, 'loop_mod.Loop']) -> None:
        """Remove the indicated tag or loop."""

        # If they specify the specific loop to delete, go ahead and delete it
        if isinstance(item, loop_mod.Loop):
            del self.loops[self.loops.index(item)]
            return

        # See if the result of get(item) is a loop. If so, delete it
        # (calls this method recursively)
        to_delete = self.__getitem__(item)
        if isinstance(to_delete, loop_mod.Loop):
            del self.loops[self.loops.index(to_delete)]
            return

        # It must be a tag. Try to delete the tag
        else:
            self.delete_tag(item)

    def __eq__(self, other) -> bool:
        """Returns True if this saveframe is equal to another saveframe,
        False if it is equal."""

        if not isinstance(other, (Saveframe, str)):
            return False

        return len(self.compare(other)) == 0

    def __getitem__(self, item: Union[int, str]) -> Union[list, 'loop_mod.Loop']:
        """Get the indicated loop or tag."""

        try:
            return self.loops[item]
        except TypeError:
            results = self.get_tag(item)
            if results:
                return results
            else:
                try:
                    return self.loop_dict[item.lower()]
                except KeyError:
                    raise KeyError("No tag or loop matching '%s'" % item)

    def __len__(self) -> int:
        """Return the number of loops in this saveframe."""

        return len(self.loops)

    def __lt__(self, other) -> bool:
        """Returns True if this saveframe sorts lower than the compared
        saveframe, false otherwise. The alphabetical ordering of the
        saveframe category is used to perform the comparison."""

        if not isinstance(other, Saveframe):
            return NotImplemented

        return self.tag_prefix < other.tag_prefix

    def __init__(self, **kwargs) -> None:
        """Don't use this directly. Use the class methods to construct:
             Saveframe.from_scratch()
             Saveframe.from_string()
             Saveframe.from_template()
             Saveframe.from_file()
             Saveframe.from_json()"""

        # They initialized us wrong
        if len(kwargs) == 0:
            raise ValueError("You should not directly instantiate a Saveframe using this method. Instead use the class"
                             " methods: Saveframe.from_scratch(), Saveframe.from_string(), Saveframe.from_template(), "
                             "Saveframe.from_file(), and Saveframe.from_json().")

        # Initialize our local variables
        self.tags: List[Any] = []
        self.loops: List[loop_mod.Loop] = []
        self._name: str = ""
        self.source: str = "unknown"
        self.category: Optional[str] = None
        self.tag_prefix: Optional[str] = None

        star_buffer: StringIO = StringIO('')

        # Update our source if it provided
        if 'source' in kwargs:
            self.source = kwargs['source']

        if 'the_string' in kwargs:
            # Parse from a string by wrapping it in StringIO
            star_buffer = StringIO(kwargs['the_string'])
            self.source = "from_string()"
        elif 'file_name' in kwargs:
            star_buffer = _interpret_file(kwargs['file_name'])
            self.source = "from_file('%s')" % kwargs['file_name']
        # Creating from template (schema)
        elif 'all_tags' in kwargs:
            schema_obj = utils.get_schema(kwargs['schema'])
            schema = schema_obj.schema
            self.category = kwargs['category']

            self.name = self.category
            if 'saveframe_name' in kwargs and kwargs['saveframe_name']:
                self.name = kwargs['saveframe_name']

            # Make sure it is a valid category
            if self.category not in [x["SFCategory"] for x in schema.values()]:
                raise ValueError("The saveframe category '%s' was not found in the dictionary." % self.category)

            s = sorted(schema.values(), key=lambda _: float(_["Dictionary sequence"]))

            loops_added = []

            for item in s:
                if item["SFCategory"] == self.category:

                    # It is a tag in this saveframe
                    if item["Loopflag"] == "N":

                        ft = utils.format_tag(item["Tag"])
                        # Set the value for sf_category and sf_framecode
                        if ft == "Sf_category":
                            self.add_tag(item["Tag"], self.category)
                        elif ft == "Sf_framecode":
                            self.add_tag(item["Tag"], self.name)
                        # If the tag is the entry ID tag, set the entry ID
                        elif item["entryIdFlg"] == "Y":
                            self.add_tag(item["Tag"], kwargs['entry_id'])
                        else:
                            tag_value = None
                            if kwargs['default_values']:
                                if item['default value'] not in definitions.NULL_VALUES:
                                    tag_value = item['default value']
                            # Unconditional add
                            if kwargs['all_tags']:
                                self.add_tag(item["Tag"], tag_value)
                            # Conditional add
                            else:
                                if item["public"] != "I":
                                    self.add_tag(item["Tag"], tag_value)

                    # It is a contained loop tag
                    else:
                        cat_formatted = utils.format_category(item["Tag"])
                        if cat_formatted not in loops_added:
                            loops_added.append(cat_formatted)
                            try:
                                self.add_loop(loop_mod.Loop.from_template(cat_formatted,
                                                                          all_tags=kwargs['all_tags'],
                                                                          schema=schema_obj))
                            except ValueError:
                                pass
            return

        elif 'saveframe_name' in kwargs:
            # If they are creating from scratch, just get the saveframe name
            self.name = kwargs['saveframe_name']
            if 'tag_prefix' in kwargs:
                self.tag_prefix = utils.format_category(kwargs['tag_prefix'])
            return

        # If we are reading from a CSV file, go ahead and parse it
        if 'csv' in kwargs and kwargs['csv']:
            csv_reader_object = csv_reader(star_buffer)
            tags = next(csv_reader_object)
            values = next(csv_reader_object)
            if len(tags) != len(values):
                raise ValueError("Your CSV data is invalid. The header length does not match the data length.")
            for ordinal in range(0, len(tags)):
                self.add_tag(tags[ordinal], values[ordinal])
            return

        tmp_entry = entry_mod.Entry.from_scratch(0)

        # Load the BMRB entry from the file
        star_buffer = StringIO("data_1 " + star_buffer.read())
        parser = parser_mod.Parser(entry_to_parse_into=tmp_entry)
        parser.parse(star_buffer.read(), source=self.source, convert_data_types=kwargs.get('convert_data_types', False))

        # Copy the first parsed saveframe into ourself
        if len(tmp_entry.frame_list) > 1:
            raise ValueError("You attempted to parse one saveframe but the source you provided had more than one "
                             "saveframe. Please either parse all saveframes as an entry or only parse one saveframe. "
                             "Saveframes detected: " + str(tmp_entry.frame_list))
        self.tags = tmp_entry[0].tags
        self.loops = tmp_entry[0].loops
        self.name = tmp_entry[0].name
        self.tag_prefix = tmp_entry[0].tag_prefix

    @property
    def empty(self) -> bool:
        """ Check if the saveframe has no data. Ignore the structural tags."""

        for tag in self.tags:
            tag_lower = tag[0].lower()
            if tag_lower not in ['sf_category', 'sf_framecode', 'id', 'entry_id', 'nmr_star_version',
                                 'original_nmr_star_version']:
                if tag[1] not in definitions.NULL_VALUES:
                    return False

        for loop in self.loops:
            if not loop.empty:
                return False

        return True

    @property
    def loop_dict(self) -> Dict[str, 'loop_mod.Loop']:
        """Returns a hash of loop category -> loop."""

        res = {}
        for each_loop in self.loops:
            if each_loop.category is not None:
                res[each_loop.category.lower()] = each_loop
        return res

    @property
    def name(self) -> Any:
        """ Returns the name of the saveframe."""

        return self._name

    @name.setter
    def name(self, name):
        """ Updates the saveframe name. """

        for char in str(name):
            if char in utils.definitions.WHITESPACE:
                raise ValueError("Saveframe names can not contain whitespace characters.")
        if name in definitions.NULL_VALUES:
            raise ValueError("Cannot set the saveframe name to a null-equivalent value.")

        # Update the sf_framecode tag too
        if 'sf_framecode' in self:
            self['sf_framecode'] = name
        self._name = name

    @property
    def tag_dict(self) -> Dict[str, str]:
        """Returns a hash of (tag name).lower() -> tag value."""

        return {x[0].lower(): x[1] for x in self.tags}

    @classmethod
    def from_scratch(cls, sf_name: str, tag_prefix: str = None, source: str = "from_scratch()"):
        """Create an empty saveframe that you can programmatically add
        to. You may also pass the tag prefix as the second argument. If
        you do not pass the tag prefix it will be set the first time you
        add a tag."""

        return cls(saveframe_name=sf_name, tag_prefix=tag_prefix, source=source)

    @classmethod
    def from_file(cls, the_file: Union[str, TextIO, BinaryIO], csv: bool = False, convert_data_types: bool = False):
        """Create a saveframe by loading in a file. Specify csv=True is
        the file is a CSV file. If the_file starts with http://,
        https://, or ftp:// then we will use those protocols to attempt
        to open the file.

        Setting convert_data_types to True will automatically convert
        the data loaded from the file into the corresponding python type as
        determined by loading the standard BMRB schema. This would mean that
        all floats will be represented as decimal.Decimal objects, all integers
        will be python int objects, strings and vars will remain strings, and
        dates will become datetime.date objects. When printing str() is called
        on all objects. Other that converting uppercase "E"s in scientific
        notation floats to lowercase "e"s this should not cause any change in
        the way re-printed NMR-STAR objects are displayed."""

        return cls(file_name=the_file, csv=csv, convert_data_types=convert_data_types)

    @classmethod
    def from_json(cls, json_dict: Union[dict, str]):
        """Create a saveframe from JSON (serialized or unserialized JSON)."""

        # If they provided a string, try to load it using JSON
        if not isinstance(json_dict, dict):
            try:
                json_dict = json.loads(json_dict)
            except (TypeError, ValueError):
                raise ValueError("The JSON you provided was neither a Python "
                                 "dictionary nor a JSON string.")

        # Make sure it has the correct keys
        for check in ["name", "tag_prefix", "tags", "loops"]:
            if check not in json_dict:
                raise ValueError("The JSON you provide must be a hash and must contain the key '%s' - even if the key "
                                 "points to None." % check)

        # Create a saveframe from scratch and populate it
        ret = Saveframe.from_scratch(json_dict['name'])
        ret.tag_prefix = json_dict['tag_prefix']
        ret.category = json_dict.get('category', None)
        ret.tags = json_dict['tags']
        ret.loops = [loop_mod.Loop.from_json(x) for x in json_dict['loops']]
        ret.source = "from_json()"

        # Return the new loop
        return ret

    @classmethod
    def from_string(cls, the_string: str, csv: bool = False, convert_data_types: bool = False):
        """Create a saveframe by parsing a string. Specify csv=True is
        the string is in CSV format and not NMR-STAR format.

        Setting convert_data_types to True will automatically convert
        the data loaded from the file into the corresponding python type as
        determined by loading the standard BMRB schema. This would mean that
        all floats will be represented as decimal.Decimal objects, all integers
        will be python int objects, strings and vars will remain strings, and
        dates will become datetime.date objects. When printing str() is called
        on all objects. Other that converting uppercase "E"s in scientific
        notation floats to lowercase "e"s this should not cause any change in
        the way re-printed NMR-STAR objects are displayed."""

        return cls(the_string=the_string, csv=csv, convert_data_types=convert_data_types)

    @classmethod
    def from_template(cls, category: str, name: str = None, entry_id: Union[str, int] = None, all_tags: bool = False,
                      default_values: bool = False, schema: Schema = None):
        """ Create a saveframe that has all of the tags and loops from the
        schema present. No values will be assigned. Specify the category
        when calling this method. Optionally also provide the name of the
        saveframe as the 'name' argument.

        The optional argument 'all_tags' forces all tags to be included
        rather than just the mandatory tags.

        The optional argument 'default_values' will insert the default
        values from the schema."""

        schema = utils.get_schema(schema)
        return cls(category=category, saveframe_name=name, entry_id=entry_id,
                   all_tags=all_tags, default_values=default_values, schema=schema,
                   source="from_template(%s)" % schema.version)

    def __repr__(self) -> str:
        """Returns a description of the saveframe."""

        return "<pynmrstar.Saveframe '%s'>" % self.name

    def __setitem__(self, key: Union[str, int], item: Union[str, 'loop_mod.Loop']) -> None:
        """Set the indicated loop or tag."""

        # It's a loop
        if isinstance(item, loop_mod.Loop):
            try:
                integer = int(str(key))
                self.loops[integer] = item
            except ValueError:
                if key.lower() in self.loop_dict:
                    for pos, tmp_loop in enumerate(self.loops):
                        if tmp_loop.category.lower() == key.lower():
                            self.loops[pos] = item
                else:
                    raise KeyError("Loop with category '%s' does not exist and therefore cannot be written to. Use "
                                   "add_loop instead." % key)
        else:
            # If the tag already exists, set its value
            self.add_tag(key, item, update=True)

    def __str__(self, first_in_category: bool = True, skip_empty_loops: bool = False,
                skip_empty_tags: bool = False, show_comments: bool = True) -> str:
        """Returns the saveframe in STAR format as a string."""

        if self.tag_prefix is None:
            raise ValueError("The tag prefix was never set!")

        # Make sure this isn't a dummy saveframe before proceeding
        try:
            width = max([len(self.tag_prefix + "." + x[0]) for x in self.tags])
        except ValueError:
            return "\nsave_%s\n\nsave_\n" % self.name

        ret_string = ""

        # Insert the comment if not disabled
        if show_comments:
            if self.category in _get_comments():
                this_comment = _get_comments()[self.category]
                if first_in_category or this_comment['every_flag']:
                    ret_string = _get_comments()[self.category]['comment']

        # Print the saveframe
        ret_string += "save_%s\n" % self.name
        pstring = "   %%-%ds  %%s\n" % width
        mstring = "   %%-%ds\n;\n%%s;\n" % width

        # Print the tags
        for each_tag in self.tags:
            if skip_empty_tags and each_tag[1] in definitions.NULL_VALUES:
                continue
            try:
                clean_tag = utils.quote_value(each_tag[1])
            except ValueError:
                raise FormattingError('Cannot generate NMR-STAR for entry, as empty strings are not valid tag values '
                                      'in NMR-STAR. Please either replace the empty strings with None objects, '
                                      'or set pynmrstar.definitions.STR_CONVERSION_DICT[\'\'] = None. '
                                      f'Saveframe: {self.name} Tag: {each_tag[0]}')

            formatted_tag = self.tag_prefix + "." + each_tag[0]
            if "\n" in clean_tag:
                ret_string += mstring % (formatted_tag, clean_tag)
            else:
                ret_string += pstring % (formatted_tag, clean_tag)

        # Print any loops
        for each_loop in self.loops:
            ret_string += each_loop.format(skip_empty_loops=skip_empty_loops, skip_empty_tags=skip_empty_tags)

        # Close the saveframe
        ret_string += "\nsave_\n"
        return ret_string

    def add_loop(self, loop_to_add: 'loop_mod.Loop') -> None:
        """Add a loop to the saveframe loops."""

        if (loop_to_add.category in self.loop_dict or
                str(loop_to_add.category).lower() in self.loop_dict):
            if loop_to_add.category is None:
                raise ValueError("You cannot have two loops with the same category in one saveframe. You are getting "
                                 "this error because you haven't yet set your loop categories.")
            else:
                raise ValueError("You cannot have two loops with the same category in one saveframe. Category: '%s'." %
                                 loop_to_add.category)

        self.loops.append(loop_to_add)

    def add_tag(self, name: str, value: Any, update: bool = False, convert_data_types: bool = False,
                line_num: Optional[int] = None) -> None:
        """Add a tag to the tag list. Does a bit of validation and
        parsing.

        Set update to True to update a tag if it exists rather
        than raise an exception.

        Set convert_data_types to True to convert the tag value from str to
        whatever type the tag is as defined in the schema. """

        if not isinstance(name, str):
            raise IllegalActionError('Tag names must be strings.', line_num=line_num)

        if "." in name:
            if name[0] != ".":
                prefix = utils.format_category(name)
                if self.tag_prefix is None:
                    self.tag_prefix = prefix
                elif self.tag_prefix != prefix:
                    raise IllegalActionError(
                        "One saveframe cannot have tags with different categories (or tags that don't "
                        f"match the set category)! '{self.tag_prefix}' vs '{prefix}'.", line_num=line_num)
                name = name[name.index(".") + 1:]
            else:
                name = name[1:]

        if name in definitions.NULL_VALUES:
            raise IllegalActionError('Cannot use a null-equivalent value as a tag name.', line_num=line_num)
        if "." in name:
            raise IllegalActionError("There cannot be more than one '.' in a tag name.", line_num=line_num)
        for char in name:
            if char in utils.definitions.WHITESPACE:
                raise IllegalActionError("Tag names can not contain whitespace characters.", line_num=line_num)

        # No duplicate tags
        if self.get_tag(name):
            if not update:
                raise IllegalActionError(f"There is already a tag with the name '{name}'. Set update=True if"
                                         f" you want to override its value.", line_num=line_num)
            else:
                tag_name_lower = name.lower()
                if tag_name_lower == "sf_category":
                    self.category = value
                if tag_name_lower == "sf_framecode":
                    if value in definitions.NULL_VALUES:
                        raise IllegalActionError("Cannot set the saveframe name to a null-equivalent value.")
                    self._name = value
                self.get_tag(name, whole_tag=True)[0][1] = value
                return

        # See if we need to convert the data type
        if convert_data_types:
            new_tag = [name, utils.get_schema().convert_tag(
                self.tag_prefix + "." + name, value, line_num=line_num)]
        else:
            new_tag = [name, value]

        # Set the category if the tag we are loading is the category
        tag_name_lower = name.lower()
        if tag_name_lower == "sf_category":
            self.category = value
        if tag_name_lower == "sf_framecode":
            if not self.name:
                self.name = value
            elif self.name != value:
                raise IllegalActionError('The Sf_framecode tag cannot be different from the saveframe name.',
                                         line_num=line_num)

        if line_num:
            new_tag.append(line_num)

        self.tags.append(new_tag)

    def add_tags(self, tag_list: list, update: bool = False) -> None:
        """Adds multiple tags to the list. Input should be a list of
        tuples that are either [key, value] or [key]. In the latter case
        the value will be set to ".".  Set update to true to update a
        tag if it exists rather than raise an exception."""

        for tag_pair in tag_list:
            if len(tag_pair) == 2:
                self.add_tag(tag_pair[0], tag_pair[1], update=update)
            elif len(tag_pair) == 1:
                self.add_tag(tag_pair[0], ".", update=update)
            else:
                raise ValueError("You provided an invalid tag/value to add: '%s'." % tag_pair)

    def add_missing_tags(self, schema: 'Schema' = None, all_tags: bool = False,
                         recursive: bool = True) -> None:
        """ Automatically adds any missing tags (according to the schema)
        and sorts the tags.

        Set recursive to False to only operate on the tags in this saveframe,
        and not those in child loops."""

        if not self.tag_prefix:
            raise ValueError("You must first specify the tag prefix of this Saveframe before calling this method. "
                             "You can do this by adding a fully qualified tag (i.e. _Entry.Sf_framecode), "
                             "by specifying the tag_prefix when calling from_scratch() or by modifying the .tag_prefix "
                             "attribute.")

        schema = utils.get_schema(schema)
        tag_prefix: str = self.tag_prefix.lower() + '.'

        for item in schema.schema_order:

            # The tag is in the loop
            if item.lower().startswith(tag_prefix):

                try:
                    # Unconditional add
                    if all_tags:
                        self.add_tag(item, None)
                    # Conditional add
                    else:
                        if schema.schema[item.lower()]["public"] != "I":
                            self.add_tag(item, None)
                except ValueError:
                    pass

        if recursive:
            for loop in self.loops:
                try:
                    loop.add_missing_tags(schema=schema, all_tags=all_tags)
                except ValueError:
                    pass

        self.sort_tags()

    def compare(self, other) -> List[str]:
        """Returns the differences between two saveframes as a list.
        Non-equal saveframes will always be detected, but specific
        differences detected depends on order of saveframes."""

        diffs = []

        # Check if this is literally the same object
        if self is other:
            return []
        # Check if the other object is our string representation
        if isinstance(other, str):
            if str(self) == other:
                return []
            else:
                return ['String was not exactly equal to saveframe.']
        elif not isinstance(other, Saveframe):
            return ['Other object is not of class Saveframe.']

        # We need to do this in case of an extra "\n" on the end of one tag
        if str(other) == str(self):
            return []

        # Do STAR comparison
        try:
            if str(self.name) != str(other.name):
                # No point comparing apples to oranges. If the tags are
                #  this different just return
                diffs.append("\tSaveframe names do not match: '%s' vs '%s'." % (self.name, other.name))
                return diffs

            if str(self.tag_prefix) != str(other.tag_prefix):
                # No point comparing apples to oranges. If the tags are
                #  this different just return
                diffs.append("\tTag prefix does not match: '%s' vs '%s'." % (self.tag_prefix, other.tag_prefix))
                return diffs

            if len(self.tags) < len(other.tags):
                diffs.append("\tNumber of tags does not match: '%d' vs '%d'. The compared entry has at least one "
                             "tag this entry does not." % (len(self.tags), len(other.tags)))

            for tag in self.tags:
                other_tag = other.get_tag(tag[0])

                if not other_tag:
                    diffs.append("\tNo tag with name '%s.%s' in compared entry." % (self.tag_prefix, tag[0]))
                    continue

                # Compare the string version of the tags in case there are
                #  non-string types. Use the conversion dict to get to str
                if (str(definitions.STR_CONVERSION_DICT.get(tag[1], tag[1])) !=
                        str(definitions.STR_CONVERSION_DICT.get(other_tag[0], other_tag[0]))):
                    diffs.append("\tMismatched tag values for tag '%s.%s': '%s' vs '%s'." %
                                 (self.tag_prefix, tag[0], str(tag[1]).replace("\n", "\\n"),
                                  str(other_tag[0]).replace("\n", "\\n")))

            if len(self.loops) != len(other.loops):
                diffs.append("\tNumber of children loops does not match: '%d' vs '%d'." %
                             (len(self.loops), len(other.loops)))

            compare_loop_dict = other.loop_dict
            for each_loop in self.loops:
                if each_loop.category.lower() in compare_loop_dict:
                    compare = each_loop.compare(compare_loop_dict[each_loop.category.lower()])
                    if len(compare) > 0:
                        diffs.append("\tLoops do not match: '%s'." % each_loop.category)
                        diffs.extend(compare)
                else:
                    diffs.append("\tNo loop with category '%s' in other entry." % each_loop.category)

        except AttributeError as err:
            diffs.append("\tAn exception occurred while comparing: '%s'." % err)

        return diffs

    def delete_tag(self, tag: str) -> None:
        """Deletes a tag from the saveframe based on tag name."""

        tag = utils.format_tag(tag).lower()

        for position, each_tag in enumerate(self.tags):
            # If the tag is a match, remove it
            if each_tag[0].lower() == tag:
                return self.tags.pop(position)

        raise KeyError("There is no tag with name '%s' to remove." % tag)

    def get_data_as_csv(self, header: bool = True, show_category: bool = True) -> str:
        """Return the data contained in the loops, properly CSVd, as a
        string. Set header to False omit the header. Set show_category
        to False to omit the loop category from the headers."""

        csv_buffer = StringIO()
        csv_writer_object = csv_writer(csv_buffer)

        if header:
            if show_category:
                csv_writer_object.writerow([str(self.tag_prefix) + "." + str(x[0]) for x in self.tags])
            else:
                csv_writer_object.writerow([str(x[0]) for x in self.tags])

        data = []
        for each_tag in self.tags:
            data.append(each_tag[1])

        csv_writer_object.writerow(data)

        csv_buffer.seek(0)
        return csv_buffer.read().replace('\r\n', '\n')

    def format(self, skip_empty_loops: bool = True, skip_empty_tags: bool = False, show_comments: bool = True) -> str:
        """ The same as calling str(Saveframe), except that you can pass options
        to customize how the saveframe is printed.

        skip_empty_loops will omit printing loops with no tags at all. (A loop with null tags is not "empty".)
        skip_empty_tags will omit tags in the saveframe and child loops which have no non-null values.
        show_comments will show the standard comments before a saveframe."""

        return self.__str__(skip_empty_loops=skip_empty_loops, show_comments=show_comments,
                            skip_empty_tags=skip_empty_tags)

    def get_json(self, serialize: bool = True) -> Union[dict, str]:
        """ Returns the saveframe in JSON format. If serialize is set to
        False a dictionary representation of the saveframe that is
        serializeable is returned."""

        saveframe_data = {
            "name": self.name,
            "category": self.category,
            "tag_prefix": self.tag_prefix,
            "tags": [[x[0], x[1]] for x in self.tags],
            "loops": [x.get_json(serialize=False) for x in self.loops]
        }

        if serialize:
            return json.dumps(saveframe_data, default=_json_serialize)
        else:
            return saveframe_data

    def get_loop_by_category(self, name: str) -> 'loop_mod.Loop':
        """Return a loop based on the loop name (category)."""

        name = utils.format_category(name).lower()
        for each_loop in self.loops:
            if str(each_loop.category).lower() == name:
                return each_loop
        raise KeyError("No loop with category '%s'." % name)

    def get_tag(self, query: str, whole_tag: bool = False) -> list:
        """Allows fetching the value of a tag by tag name. Returns
        a list of all matching tag values.

        Specify whole_tag=True and the [tag_name, tag_value] pair will be
        returned instead of just the value"""

        results = []

        # Make sure this is the correct saveframe if they specify a tag
        #  prefix
        if "." in query:
            tag_prefix = utils.format_category(query)
        else:
            tag_prefix = self.tag_prefix

        # Check the loops
        for each_loop in self.loops:
            if ((each_loop.category is not None and tag_prefix is not None and
                 each_loop.category.lower() == tag_prefix.lower())):
                results.extend(each_loop.get_tag(query, whole_tag=whole_tag))

        # Check our tags
        query = utils.format_tag(query).lower()
        if tag_prefix is not None and tag_prefix.lower() == self.tag_prefix.lower():
            for tag in self.tags:
                if query == tag[0].lower():
                    if whole_tag:
                        results.append(tag)
                    else:
                        results.append(tag[1])

        return results

    def loop_iterator(self) -> Iterable['loop_mod.Loop']:
        """Returns an iterator for saveframe loops."""

        return iter(self.loops)

    def print_tree(self) -> None:
        """Prints a summary, tree style, of the loops in the saveframe."""

        print(repr(self))
        for pos, each_loop in enumerate(self):
            print("\t[%d] %s" % (pos, repr(each_loop)))

    def set_tag_prefix(self, tag_prefix: str) -> None:
        """Set the tag prefix for this saveframe."""

        self.tag_prefix = utils.format_category(tag_prefix)

    def sort_tags(self, schema: Schema = None) -> None:
        """ Sort the tags so they are in the same order as a BMRB
        schema. Will automatically use the standard schema if none
        is provided."""

        schema = utils.get_schema(schema)

        def sort_key(x) -> int:
            return schema.tag_key(self.tag_prefix + "." + x[0])

        self.tags.sort(key=sort_key)

    def tag_iterator(self) -> Iterable[list]:
        """Returns an iterator for saveframe tags."""

        return iter(self.tags)

    def validate(self, validate_schema: bool = True, schema: Schema = None, validate_star: bool = True):
        """Validate a saveframe in a variety of ways. Returns a list of
        errors found. 0-length list indicates no errors found. By
        default all validation modes are enabled.

        validate_schema - Determines if the entry is validated against
        the NMR-STAR schema. You can pass your own custom schema if desired,
        otherwise the schema will be fetched from the BMRB servers.

        validate_star - Determines if the STAR syntax checks are ran."""

        errors = []

        my_category = self.category
        if not my_category:
            errors.append("Cannot properly validate saveframe: '%s'. No saveframe category defined." % self.name)
            my_category = None

        if validate_schema:
            # Get the default schema if we are not passed a schema
            my_schema = utils.get_schema(schema)

            for tag in self.tags:
                line_number = str(tag[2]) + " of original file" if len(tag) > 2 else None
                formatted_tag = self.tag_prefix + "." + tag[0]
                cur_errors = my_schema.val_type(formatted_tag, tag[1], category=my_category, line_number=line_number)
                errors.extend(cur_errors)

        # Check the loops for errors
        for each_loop in self.loops:
            errors.extend(each_loop.validate(validate_schema=validate_schema, schema=schema,
                                             validate_star=validate_star, category=my_category))

        return errors

    def write_to_file(self, file_name: str, format_: str = "nmrstar", show_comments: bool = True,
                      skip_empty_loops: bool = False, skip_empty_tags: bool = False):
        """ Writes the saveframe to the specified file in NMR-STAR format.

        Optionally specify:
        show_comments=False to disable the comments that are by default inserted. Ignored when writing json.
        skip_empty_loops=False to force printing loops with no tags at all (loops with null tags are still printed)
        skip_empty_tags=True will omit tags in the saveframes and loops which have no non-null values.
        format_=json to write to the file in JSON format."""

        if format_ not in ["nmrstar", "json"]:
            raise ValueError("Invalid output format.")

        data_to_write = ''
        if format_ == "nmrstar":
            data_to_write = self.format(show_comments=show_comments, skip_empty_loops=skip_empty_loops,
                                        skip_empty_tags=skip_empty_tags)
        elif format_ == "json":
            data_to_write = self.get_json()

        out_file = open(file_name, "w")
        out_file.write(data_to_write)
        out_file.close()



class ParsingError(ValueError):
    """ Something went wrong when parsing. """

    def __init__(self, message, line_number: int = None):
        Exception.__init__(self)
        self.message = message
        self.line_number = line_number

    def __repr__(self) -> str:
        if self.line_number is not None:
            return 'ParsingError("%s") on line %d' % (self.message, self.line_number)
        else:
            return 'ParsingError("%s")' % self.message

    def __str__(self) -> str:
        if self.line_number is not None:
            return "%s on line %d" % (self.message, self.line_number)
        else:
            return self.message


class FormattingError(Exception):
    """ Something went wrong when formatting a file. """

    def __init__(self, message):
        Exception.__init__(self)
        self.message = message

    def __repr__(self) -> str:
        return 'FormattingError("%s")' % self.message

    def __str__(self) -> str:
        return self.message


class IllegalActionError(ValueError):
    """ The user tried to do something illegal.

    Also supports intelligently formatting errors to show the line number
    if the issue is in a file. """

    def __init__(self, message, line_num=None):
        Exception.__init__(self)
        if line_num:
            self.message = message + f" Error in parsed data on line number: {line_num}"
        else:
            self.message = message

    def __repr__(self) -> str:
        return 'ModificationError("%s")' % self.message

    def __str__(self) -> str:
        return self.message



import logging
import re
from typing import Optional, Any

from pynmrstar import definitions, entry as entry_mod, loop as loop_mod, saveframe as saveframe_mod
from pynmrstar._internal import _get_cnmrstar
from pynmrstar.exceptions import ParsingError, IllegalActionError

cnmrstar = _get_cnmrstar()


class Parser(object):
    """Parses an entry. You should not ever use this class directly."""

    def __init__(self, entry_to_parse_into: 'entry_mod.Entry' = None) -> None:

        # Just make an entry to parse into if called with no entry passed
        if entry_to_parse_into is None:
            entry_to_parse_into = entry_mod.Entry.from_scratch("")

        self.ent: entry_mod.Entry = entry_to_parse_into
        self.to_process: str = ""
        self.full_data: str = ""
        self.index: int = 0
        self.token: str = ""
        self.source: str = "unknown"
        self.delimiter: str = " "
        self.line_number: int = 0

    def get_line_number(self) -> int:
        """ Returns the current line number that is in the process of
        being parsed."""

        if cnmrstar is not None:
            return self.line_number
        else:
            return self.full_data[0:self.index].count("\n") + 1

    def get_token(self, raise_parse_warnings: bool = False) -> str:
        """ Returns the next token in the parsing process."""

        if cnmrstar is not None:
            try:
                self.token, self.line_number, self.delimiter = cnmrstar.get_token_full()
            except ValueError as err:
                raise ParsingError(str(err))
        else:
            self.real_get_token(raise_parse_warnings)
            self.line_number = 0

            if self.delimiter == ";":
                try:
                    # Un-indent value which contain STAR multi-line values
                    # Only do this check if we are comma-delineated
                    if self.token.startswith("\n   "):
                        # Only remove the whitespaces if all lines have them
                        trim = True
                        for pos in range(1, len(self.token) - 4):
                            if self.token[pos] == "\n":
                                if self.token[pos + 1:pos + 4] != "   ":
                                    trim = False

                        if trim and "\n   ;" in self.token:
                            self.token = self.token[:-1].replace("\n   ", "\n")

                except AttributeError:
                    pass

        # Return the token
        return self.token

    @staticmethod
    def index_handle(haystack: Any, needle: Any, start_pos: Optional[int] = None) -> Optional[int]:
        """ Finds the index while catching ValueError and returning
        None instead."""

        try:
            return haystack.index(needle, start_pos)
        except ValueError:
            return None

    @staticmethod
    def next_whitespace(data) -> int:
        """ Returns the position of the next whitespace character in the
        provided string. If no whitespace it returns the length of the
        string."""

        for pos, char in enumerate(data):
            if char in definitions.WHITESPACE:
                return pos
        return len(data)

    def load_data(self, data: str) -> None:
        """ Loads data in preparation of parsing and cleans up newlines
        and massages the data to make parsing work properly when multi-line
        values aren't as expected. Useful for manually getting tokens from
        the parser."""

        # Fix DOS line endings
        data = data.replace("\r\n", "\n").replace("\r", "\n")

        # Change '\n; data ' started multi-lines to '\n;\ndata'
        data = re.sub(r'\n;([^\n]+?)\n', r'\n;\n\1\n', data)

        if cnmrstar is not None:
            cnmrstar.load_string(data)
        else:
            self.full_data = data + "\n"

    def parse(self, data: str, source: str = "unknown", raise_parse_warnings: bool = False,
              convert_data_types: bool = False) -> 'entry_mod.Entry':
        """ Parses the string provided as data as an NMR-STAR entry
        and returns the parsed entry. Raises ParsingError on exceptions.

        Set raise_parse_warnings to raise an exception if the file has
        something technically incorrect, but still parsable.

        Following is a list of the types of errors that would trigger
        raise_parse_warnings:

        * A loop with no data was found.
        * A loop with no tags or values was found.
        * A tag with an improper multi-line value was found.
        Multi-line values should look like this:
        \n;\nThe multi-line\nvalue here.\n;\n
        but the tag looked like this:
        \n; The multi-line\nvalue here.\n;\n"""

        # Prepare the data for parsing
        self.load_data(data)

        # Create the NMRSTAR object
        cur_data = []

        # Get the first token
        self.get_token()

        # Make sure this is actually a STAR file
        if not self.token.lower().startswith("data_"):
            raise ParsingError("Invalid file. NMR-STAR files must start with 'data_'. Did you accidentally select the "
                               "wrong file?", self.get_line_number())

        # Make sure there is a data name
        elif len(self.token) < 6:
            raise ParsingError("'data_' must be followed by data name. Simply 'data_' is not allowed.",
                               self.get_line_number())

        if self.delimiter != " ":
            raise ParsingError("The data_ keyword may not be quoted or semicolon-delimited.")

        # Set the entry_id
        self.ent._entry_id = self.token[5:]
        self.source = source

        # We are expecting to get saveframes
        while self.get_token() is not None:

            if not self.token.lower().startswith("save_"):
                raise ParsingError("Only 'save_NAME' is valid in the body of a NMR-STAR file. Found '%s'." % self.token,
                                   self.get_line_number())

            if len(self.token) < 6:
                raise ParsingError("'save_' must be followed by saveframe name. You have a 'save_' tag which is "
                                   "illegal without a specified saveframe name.", self.get_line_number())

            if self.delimiter != " ":
                raise ParsingError("The save_ keyword may not be quoted or semicolon-delimited.",
                                   self.get_line_number())

            # Add the saveframe
            cur_frame: Optional[saveframe_mod.Saveframe] = saveframe_mod.Saveframe.from_scratch(self.token[5:],
                                                                                                source=source)
            self.ent.add_saveframe(cur_frame)

            # We are in a saveframe
            while self.get_token() is not None:

                if self.token.lower() == "loop_":
                    if self.delimiter != " ":
                        raise ParsingError("The loop_ keyword may not be quoted or semicolon-delimited.",
                                           self.get_line_number())

                    cur_loop: Optional[loop_mod.Loop] = loop_mod.Loop.from_scratch(source=source)

                    # We are in a loop
                    seen_data = False
                    in_loop = True
                    while in_loop and self.get_token() is not None:

                        # Add a tag
                        if self.token.startswith("_"):
                            if self.delimiter != " ":
                                raise ParsingError("Loop tags may not be quoted or semicolon-delimited.",
                                                   self.get_line_number())
                            if seen_data:
                                raise ParsingError("Cannot have more loop tags after loop data.")
                            cur_loop.add_tag(self.token)

                        # On to data
                        else:

                            # Now that we have the tags we can add the loop
                            #  to the current saveframe
                            cur_frame.add_loop(cur_loop)

                            # We are in the data block of a loop
                            while self.token is not None:
                                if self.token.lower() == "stop_":
                                    if self.delimiter != " ":
                                        raise ParsingError(
                                            "The stop_ keyword may not be quoted or semicolon-delimited.",
                                            self.get_line_number())
                                    if len(cur_loop.tags) == 0:
                                        if raise_parse_warnings:
                                            raise ParsingError("Loop with no tags.", self.get_line_number())
                                        else:
                                            logging.warning('Loop with no tags in parsed file on line: %s' %
                                                            self.get_line_number())
                                        cur_loop = None
                                    if not seen_data:
                                        if raise_parse_warnings:
                                            raise ParsingError("Loop with no data.", self.get_line_number())
                                        else:
                                            logging.warning("Loop with no data on line: %s" % self.get_line_number())

                                    if len(cur_data) > 0:
                                        try:
                                            cur_loop.add_data(cur_data, rearrange=True,
                                                              convert_data_types=convert_data_types)
                                        # If there is an issue with the loops during parsing, raise a parse error
                                        #  rather than the ValueError that would be raised if they made the mistake
                                        #   directly
                                        except ValueError as e:
                                            raise ParsingError(str(e))
                                    cur_data = []

                                    cur_loop = None
                                    in_loop = False
                                    break
                                else:
                                    if len(cur_loop.tags) == 0:
                                        raise ParsingError("Data found in loop before loop tags.",
                                                           self.get_line_number())

                                    if (self.token in definitions.RESERVED_KEYWORDS and
                                            self.delimiter == " "):
                                        raise ParsingError("Cannot use keywords as data values unless quoted or "
                                                           "semi-colon delimited. Perhaps this is a loop that wasn't "
                                                           "properly terminated? Illegal value: " + self.token,
                                                           self.get_line_number())
                                    cur_data.append(self.token)
                                    seen_data = True

                                # Get the next token
                                self.get_token()

                    if not self.token or self.token.lower() != "stop_":
                        raise ParsingError("Loop improperly terminated at end of file.", self.get_line_number())

                # Close saveframe
                elif self.token.lower() == "save_":
                    if self.delimiter not in " ;":
                        raise ParsingError("The save_ keyword may not be quoted or semicolon-delimited.",
                                           self.get_line_number())

                    if cur_frame.tag_prefix is None:
                        raise ParsingError("The tag prefix was never set! Either the saveframe had no tags, you "
                                           "tried to read a version 2.1 file, or there is something else wrong with "
                                           "your file. Saveframe error occurred within: '%s'" % cur_frame.name)
                    break

                # Invalid content in saveframe
                elif not self.token.startswith("_"):
                    raise ParsingError("Invalid token found in saveframe '%s': '%s'" % (cur_frame.name, self.token),
                                       self.get_line_number())

                # Add a tag
                else:
                    if self.delimiter != " ":
                        raise ParsingError("Saveframe tags may not be quoted or semicolon-delimited.",
                                           self.get_line_number())
                    cur_tag: Optional[str] = self.token

                    # We are in a saveframe and waiting for the saveframe tag
                    self.get_token()
                    if self.delimiter == " ":
                        if self.token in definitions.RESERVED_KEYWORDS:
                            raise ParsingError("Cannot use keywords as data values unless quoted or semi-colon "
                                               "delimited. Illegal value: " +
                                               self.token, self.get_line_number())
                        if self.token.startswith("_"):
                            raise ParsingError(
                                "Cannot have a tag value start with an underscore unless the entire value "
                                "is quoted. You may be missing a data value on the previous line. "
                                "Illegal value: " + self.token, self.get_line_number())
                    try:
                        cur_frame.add_tag(cur_tag, self.token, line_num=self.get_line_number(),
                                          convert_data_types=convert_data_types)
                    except IllegalActionError as err:
                        raise ParsingError(str(err))

            if not self.token or self.token.lower() != "save_":
                raise ParsingError("Saveframe improperly terminated at end of file.", self.get_line_number())

        # Free the memory of the original copy of the data we parsed
        self.full_data = None

        # Reset the parser
        if cnmrstar is not None:
            cnmrstar.reset()

        return self.ent

    def real_get_token(self, raise_parse_warnings: bool = False) -> Optional[str]:
        """ Actually processes the input data to find a token. get_token
        is just a wrapper around this with some exception handling."""

        # Reset the delimiter
        self.delimiter = " "

        # Nothing left
        if self.token is None:
            return

        # We're at the end if the index is the length
        if self.index == len(self.full_data):
            self.token = None
            return

        # Get just a single line of the file
        raw_tmp = ""
        tmp = ""
        while len(tmp) == 0:
            self.index += len(raw_tmp)

            try:
                newline_index = self.full_data.index("\n", self.index + 1)
                raw_tmp = self.full_data[self.index:newline_index]
            except ValueError:
                # End of file
                self.token = self.full_data[self.index:].lstrip(definitions.WHITESPACE)
                if self.token == "":
                    self.token = None
                self.index = len(self.full_data)
                return

            newline_index = self.full_data.index("\n", self.index + 1)
            raw_tmp = self.full_data[self.index:newline_index + 1]
            tmp = raw_tmp.lstrip(definitions.WHITESPACE)

        # If it is a multi-line comment, recalculate our viewing window
        if tmp[0:2] == ";\n":
            try:
                q_start = self.full_data.index(";\n", self.index)
                q_end = self.full_data.index("\n;", q_start) + 3
            except ValueError:
                q_end = len(self.full_data)

            raw_tmp = self.full_data[self.index:q_end]
            tmp = raw_tmp.lstrip()

        self.index += len(raw_tmp) - len(tmp)

        # Skip comments
        if tmp.startswith("#"):
            self.index += len(tmp)
            return self.get_token()

        # Handle multi-line values
        if tmp.startswith(";\n"):
            tmp = tmp[2:]

            # Search for end of multi-line value
            if "\n;" in tmp:
                until = tmp.index("\n;")
                valid = self.index_handle(tmp, "\n;\n")

                # The line is terminated properly
                if valid == until:
                    self.token = tmp[0:until + 1]
                    self.index += until + 4
                    self.delimiter = ";"
                    return

                # The line was terminated improperly
                else:
                    if self.next_whitespace(tmp[until + 2:]) == 0:
                        if raise_parse_warnings:
                            raise ParsingError("Warning: Technically invalid line found in file. Multi-line values "
                                               "should terminate with \\n;\\n but in this file only \\n; with "
                                               "non-return whitespace following was found.", self.get_line_number())
                        else:
                            logging.warning("Technically invalid line found in file. Multi-line values "
                                            "should terminate with \\n;\\n but in this file only \\n; with non-return "
                                            "whitespace following was found. Line: %s" % self.get_line_number())
                        self.token = tmp[0:until + 1]
                        self.index += until + 4
                        self.delimiter = ";"
                        return
                    else:
                        raise ParsingError('Invalid file. A multi-line value ended with a "\\n;" and then a '
                                           'non-whitespace value. Multi-line values should end with "\\n;\\n".',
                                           self.get_line_number())
            else:
                raise ParsingError("Invalid file. Multi-line comment never ends. Multi-line comments must terminate "
                                   "with a line that consists ONLY of a ';' without characters before or after. (Other "
                                   "than the newline.)", self.get_line_number())

        # Handle values quoted with '
        if tmp.startswith("'"):
            until = self.index_handle(tmp, "'", 1)

            if until is None:
                raise ParsingError("Invalid file. Single quoted value was never terminated.", self.get_line_number())

            # Make sure we don't stop for quotes that are not followed by whitespace
            try:
                while tmp[until + 1:until + 2] not in definitions.WHITESPACE:
                    until = self.index_handle(tmp, "'", until + 1)
            except TypeError:
                raise ParsingError("Invalid file. Single quoted value was never terminated.", self.get_line_number())

            self.token = tmp[1:until]
            self.index += until + 1
            self.delimiter = "'"
            return

        # Handle values quoted with "
        if tmp.startswith('"'):
            until = self.index_handle(tmp, '"', 1)

            if until is None:
                raise ParsingError("Invalid file. Double quoted value was never terminated.", self.get_line_number())

            # Make sure we don't stop for quotes that are not followed by whitespace
            try:
                while tmp[until + 1:until + 2] not in definitions.WHITESPACE:
                    until = self.index_handle(tmp, '"', until + 1)
            except TypeError:
                raise ParsingError("Invalid file. Double quoted value was never terminated.", self.get_line_number())

            self.token = tmp[1:until]
            self.index += until + 1
            self.delimiter = '"'
            return

        # Figure out where this token ends
        white = self.next_whitespace(tmp)
        if white == len(tmp):
            self.token = tmp
            self.index += len(self.token) + 1
            if self.token[0] == "$" and len(self.token) > 1:
                self.delimiter = '$'
            return

        # The token isn't anything special, just return it
        self.index += white
        self.token = tmp[0:white]
        if self.token[0] == "$" and len(self.token) > 1:
            self.delimiter = '$'
        return



#!/usr/bin/python3

""" NMR-STAR definitions and other module parameters live here. Technically
you can edit them, but you should really know what you're doing.

Adding key->value pairs to STR_CONVERSION_DICT will automatically convert tags
whose value matches "key" to the string "value" when printing. This allows you
to set the default conversion value for Booleans or other objects.

WARNING: STR_CONVERSION_DICT cannot contain both booleans and arithmetic types.
Attempting to use both will cause an issue since boolean True == 1 in python
and False == 0.
"""

NULL_VALUES = ['', ".", "?", None]
WHITESPACE: str = " \t\n\v"
RESERVED_KEYWORDS = ["data_", "save_", "loop_", "stop_", "global_"]
STR_CONVERSION_DICT: dict = {None: "."}

API_URL: str = "https://api.bmrb.io/v2"
SCHEMA_URL: str = 'https://raw.githubusercontent.com/uwbmrb/nmr-star-dictionary/master/xlschem_ann.csv'
COMMENT_URL: str = "https://raw.githubusercontent.com/uwbmrb/PyNMRSTAR/v3/reference_files/comments.str"
TYPES_URL: str = "https://raw.githubusercontent.com/uwbmrb/PyNMRSTAR/v3/pynmrstar/reference_files/data_types.csv"
FTP_URL: str = "https://bmrb.io/ftp/pub/bmrb"



import hashlib
import json
import logging
from io import StringIO
from typing import TextIO, BinaryIO, Union, List, Optional, Dict, Any

from pynmrstar import definitions, utils, loop as loop_mod, parser as parser_mod, saveframe as saveframe_mod
from pynmrstar._internal import _json_serialize, _interpret_file, _get_entry_from_database
from pynmrstar.schema import Schema


class Entry(object):
    """An OO representation of a BMRB entry. You can initialize this
    object several ways; (e.g. from a file, from the official database,
    from scratch) see the class methods below."""

    def __delitem__(self, item: Union['saveframe_mod.Saveframe', int, str]) -> None:
        """Remove the indicated saveframe."""

        if isinstance(item, saveframe_mod.Saveframe):
            del self.frame_list[self.frame_list.index(item)]
            return
        else:
            self.__delitem__(self.__getitem__(item))

    def __eq__(self, other) -> bool:
        """Returns True if this entry is equal to another entry, false
        if it is not equal."""

        if not isinstance(other, (Entry, str)):
            return False

        return len(self.compare(other)) == 0

    def __getitem__(self, item: Union[int, str]) -> 'saveframe_mod.Saveframe':
        """Get the indicated saveframe."""

        try:
            return self.frame_list[item]
        except TypeError:
            return self.get_saveframe_by_name(item)

    def __init__(self, **kwargs) -> None:
        """ You should not directly instantiate an Entry using this method.
            Instead use the class methods:"
              Entry.from_database()
              Entry.from_file()
              Entry.from_string()
              Entry.from_scratch()
              Entry.from_json()
              Entry.from_template()"""

        # Default initializations
        self._entry_id: Union[str, int] = 0
        self.frame_list: List[saveframe_mod.Saveframe] = []
        self.source: Optional[str] = None

        # They initialized us wrong
        if len(kwargs) == 0:
            raise ValueError("You should not directly instantiate an Entry using this method. Instead use the class "
                             "methods: Entry.from_database(), Entry.from_file(), Entry.from_string(), "
                             "Entry.from_scratch(), and Entry.from_json().")

        if 'the_string' in kwargs:
            # Parse from a string by wrapping it in StringIO
            star_buffer: StringIO = StringIO(kwargs['the_string'])
            self.source = "from_string()"
        elif 'file_name' in kwargs:
            star_buffer = _interpret_file(kwargs['file_name'])
            self.source = "from_file('%s')" % kwargs['file_name']
        # Creating from template (schema)
        elif 'all_tags' in kwargs:
            self._entry_id = kwargs['entry_id']

            saveframe_categories: dict = {}
            schema = utils.get_schema(kwargs['schema'])
            schema_obj = schema.schema
            for tag in [schema_obj[x.lower()] for x in schema.schema_order]:
                category = tag['SFCategory']
                if category not in saveframe_categories:
                    saveframe_categories[category] = True
                    templated_saveframe = saveframe_mod.Saveframe.from_template(category, category + "_1",
                                                                                entry_id=self._entry_id,
                                                                                all_tags=kwargs['all_tags'],
                                                                                default_values=kwargs['default_values'],
                                                                                schema=schema)
                    self.frame_list.append(templated_saveframe)
            entry_saveframe = self.get_saveframes_by_category('entry_information')[0]
            entry_saveframe['NMR_STAR_version'] = schema.version
            entry_saveframe['Original_NMR_STAR_version'] = schema.version
            return
        else:
            # Initialize a blank entry
            self._entry_id = kwargs['entry_id']
            self.source = "from_scratch()"
            return

        # Load the BMRB entry from the file
        parser: parser_mod.Parser = parser_mod.Parser(entry_to_parse_into=self)
        parser.parse(star_buffer.read(), source=self.source, convert_data_types=kwargs.get('convert_data_types', False))

    def __len__(self) -> int:
        """ Returns the number of saveframes in the entry."""

        return len(self.frame_list)

    def __repr__(self) -> str:
        """Returns a description of the entry."""

        return "<pynmrstar.Entry '%s' %s>" % (self._entry_id, self.source)

    def __setitem__(self, key: Union[int, str], item: 'saveframe_mod.Saveframe') -> None:
        """Set the indicated saveframe."""

        # It is a saveframe
        if isinstance(item, saveframe_mod.Saveframe):
            # Add by ordinal
            if isinstance(key, int):
                self.frame_list[key] = item
            else:
                # Add by key
                contains_frame: bool = False
                for pos, frame in enumerate(self.frame_list):
                    if frame.name == key:
                        if contains_frame:
                            raise ValueError('Two saveframes with the same name in Entry (%s) - this entry is invalid'
                                             ' NMR-STAR. Did you manually edit the frame_list? Please use the '
                                             'add_saveframe() method to add new saveframes.' % frame.name)
                        self.frame_list[pos] = item
                        contains_frame = True

                if not contains_frame:
                    raise KeyError("Saveframe with name '%s' does not exist and therefore cannot be written to. Use "
                                   "the add_saveframe() method to add new saveframes." % key)
        else:
            raise ValueError("You can only assign a saveframe to an entry splice.")

    def __str__(self, skip_empty_loops: bool = False, skip_empty_tags: bool = False, show_comments: bool = True) -> str:
        """Returns the entire entry in STAR format as a string."""

        sf_strings = []
        seen_saveframes = {}
        for saveframe_obj in self:
            if saveframe_obj.category in seen_saveframes:
                sf_strings.append(saveframe_obj.format(skip_empty_loops=skip_empty_loops,
                                                       skip_empty_tags=skip_empty_tags, show_comments=False))
            else:
                sf_strings.append(saveframe_obj.format(skip_empty_loops=skip_empty_loops,
                                                       skip_empty_tags=skip_empty_tags, show_comments=show_comments))
                seen_saveframes[saveframe_obj.category] = True

        return "data_%s\n\n%s" % (self.entry_id, "\n".join(sf_strings))

    @property
    def category_list(self) -> List[str]:
        """ Returns a list of the unique categories present in the entry. """

        category_list = []
        for saveframe in self.frame_list:
            category = saveframe.category
            if category and category not in category_list:
                category_list.append(category)
        return list(category_list)

    @property
    def empty(self) -> bool:
        """ Check if the entry has no data. Ignore the structural tags."""

        for saveframe in self.frame_list:
            if not saveframe.empty:
                return False

        return True

    @property
    def entry_id(self) -> Union[str, int]:
        return self._entry_id

    @entry_id.setter
    def entry_id(self, value: Union[str, int]) -> None:
        self._entry_id = value

        schema = utils.get_schema()
        for saveframe in self.frame_list:
            for tag in saveframe.tags:
                fqtn = (saveframe.tag_prefix + "." + tag[0]).lower()

                try:
                    if schema.schema[fqtn]['entryIdFlg'] == 'Y':
                        tag[1] = self._entry_id
                except KeyError:
                    pass

            for loop in saveframe.loops:
                for tag in loop.tags:
                    fqtn = (loop.category + "." + tag).lower()
                    try:
                        if schema.schema[fqtn]['entryIdFlg'] == 'Y':
                            loop[tag] = [self._entry_id] * len(loop[tag])
                    except KeyError:
                        pass

    @property
    def frame_dict(self) -> Dict[str, 'saveframe_mod.Saveframe']:
        """Returns a dictionary of saveframe name -> saveframe object"""

        fast_dict = dict((frame.name, frame) for frame in self.frame_list)

        # If there are no duplicates then continue
        if len(fast_dict) == len(self.frame_list):
            return fast_dict

        # Figure out where the duplicate is
        frame_dict = {}

        for frame in self.frame_list:
            if frame.name in frame_dict:
                raise ValueError("The entry has multiple saveframes with the same name. That is illegal. Please remove "
                                 "or rename one. Duplicate name: %s" % frame.name)
            frame_dict[frame.name] = frame

        return frame_dict

    @classmethod
    def from_database(cls, entry_num: Union[str, int], convert_data_types: bool = False):
        """Create an entry corresponding to the most up to date entry on
        the public BMRB server. (Requires ability to initiate outbound
        HTTP connections.)

        Setting convert_data_types to True will automatically convert
        the data loaded from the file into the corresponding python type as
        determined by loading the standard BMRB schema. This would mean that
        all floats will be represented as decimal.Decimal objects, all integers
        will be python int objects, strings and vars will remain strings, and
        dates will become datetime.date objects. When printing str() is called
        on all objects. Other that converting uppercase "E"s in scientific
        notation floats to lowercase "e"s this should not cause any change in
        the way re-printed NMR-STAR objects are displayed."""

        return _get_entry_from_database(entry_num, convert_data_types=convert_data_types)

    @classmethod
    def from_file(cls, the_file: Union[str, TextIO, BinaryIO], convert_data_types: bool = False):
        """Create an entry by loading in a file. If the_file starts with
        http://, https://, or ftp:// then we will use those protocols to
        attempt to open the file.
        
        Setting convert_data_types to True will automatically convert
        the data loaded from the file into the corresponding python type as
        determined by loading the standard BMRB schema. This would mean that
        all floats will be represented as decimal.Decimal objects, all integers
        will be python int objects, strings and vars will remain strings, and
        dates will become datetime.date objects. When printing str() is called
        on all objects. Other that converting uppercase "E"s in scientific
        notation floats to lowercase "e"s this should not cause any change in
        the way re-printed NMR-STAR objects are displayed."""

        return cls(file_name=the_file, convert_data_types=convert_data_types)

    @classmethod
    def from_json(cls, json_dict: Union[dict, str]):
        """Create an entry from JSON (serialized or unserialized JSON)."""

        # If they provided a string, try to load it using JSON
        if not isinstance(json_dict, dict):
            try:
                json_dict = json.loads(json_dict)
            except (TypeError, ValueError):
                raise ValueError("The JSON you provided was neither a Python dictionary nor a JSON string.")

        # Make sure it has the correct keys
        if "saveframes" not in json_dict:
            raise ValueError("The JSON you provide must be a hash and must contain the key 'saveframes' - even if the "
                             "key points to 'None'.")
        if "entry_id" not in json_dict and "bmrb_id" not in json_dict:
            raise ValueError("The JSON you provide must be a hash and must contain the key 'entry_id' - even if the"
                             " key points to 'None'.")
        # Until the migration is complete, 'bmrb_id' is a synonym for
        #  'entry_id'
        if 'entry_id' not in json_dict:
            json_dict['entry_id'] = json_dict['bmrb_id']

        # Create an entry from scratch and populate it
        ret = Entry.from_scratch(json_dict['entry_id'])
        ret.frame_list = [saveframe_mod.Saveframe.from_json(x) for x in json_dict['saveframes']]
        ret.source = "from_json()"

        # Return the new loop
        return ret

    @classmethod
    def from_string(cls, the_string: str, convert_data_types: bool = False):
        """Create an entry by parsing a string.


        Setting convert_data_types to True will automatically convert
        the data loaded from the file into the corresponding python type as
        determined by loading the standard BMRB schema. This would mean that
        all floats will be represented as decimal.Decimal objects, all integers
        will be python int objects, strings and vars will remain strings, and
        dates will become datetime.date objects. When printing str() is called
        on all objects. Other that converting uppercase "E"s in scientific
        notation floats to lowercase "e"s this should not cause any change in
        the way re-printed NMR-STAR objects are displayed."""

        return cls(the_string=the_string, convert_data_types=convert_data_types)

    @classmethod
    def from_scratch(cls, entry_id: Union[str, int]):
        """Create an empty entry that you can programmatically add to.
        You must pass a value corresponding to the Entry ID.
        (The unique identifier "xxx" from "data_xxx".)"""

        return cls(entry_id=entry_id)

    @classmethod
    def from_template(cls, entry_id, all_tags=False, default_values=False, schema=None) -> 'Entry':
        """ Create an entry that has all of the saveframes and loops from the
        schema present. No values will be assigned. Specify the entry
        ID when calling this method.

        The optional argument 'all_tags' forces all tags to be included
        rather than just the mandatory tags.

        The optional argument 'default_values' will insert the default
        values from the schema.

        The optional argument 'schema' allows providing a custom schema."""

        schema = utils.get_schema(schema)
        entry = cls(entry_id=entry_id, all_tags=all_tags, default_values=default_values, schema=schema)
        entry.source = "from_template(%s)" % schema.version
        return entry

    def add_saveframe(self, frame) -> None:
        """Add a saveframe to the entry."""

        if not isinstance(frame, saveframe_mod.Saveframe):
            raise ValueError("You can only add instances of saveframes using this method.")

        # Do not allow the addition of saveframes with the same name
        #  as a saveframe which already exists in the entry
        if frame.name in self.frame_dict:
            raise ValueError("Cannot add a saveframe with name '%s' since a saveframe with that name already exists "
                             "in the entry." % frame.name)

        self.frame_list.append(frame)

    def compare(self, other) -> List[str]:
        """Returns the differences between two entries as a list.
        Non-equal entries will always be detected, but specific differences
        detected depends on order of entries."""

        diffs = []
        if self is other:
            return []
        if isinstance(other, str):
            if str(self) == other:
                return []
            else:
                return ['String was not exactly equal to entry.']
        elif not isinstance(other, Entry):
            return ['Other object is not of class Entry.']
        try:
            if str(self.entry_id) != str(other.entry_id):
                diffs.append("Entry ID does not match between entries: '%s' vs '%s'." % (self.entry_id, other.entry_id))
            if len(self.frame_list) != len(other.frame_list):
                diffs.append("The number of saveframes in the entries are not equal: '%d' vs '%d'." %
                             (len(self.frame_list), len(other.frame_list)))
            for frame in self.frame_list:
                other_frame_dict = other.frame_dict
                if frame.name not in other_frame_dict:
                    diffs.append("No saveframe with name '%s' in other entry." % frame.name)
                else:
                    comp = frame.compare(other_frame_dict[frame.name])
                    if len(comp) > 0:
                        diffs.append("Saveframes do not match: '%s'." % frame.name)
                        diffs.extend(comp)

        except AttributeError as err:
            diffs.append("An exception occurred while comparing: '%s'." % err)

        return diffs

    def add_missing_tags(self, schema: 'Schema' = None, all_tags: bool = False) -> None:
        """ Automatically adds any missing tags (according to the schema)
        to all saveframes and loops and sorts the tags. """

        for saveframe in self.frame_list:
            saveframe.add_missing_tags(schema=schema, all_tags=all_tags)

    def delete_empty_saveframes(self) -> None:
        """ This method will delete all empty saveframes in an entry
        (the loops in the saveframe must also be empty for the saveframe
        to be deleted). "Empty" means no values in tags, not no tags present."""

        for pos, entry in enumerate(self.frame_list):
            if entry.empty:
                del self.frame_list[pos]

    def format(self, skip_empty_loops: bool = True, skip_empty_tags: bool = False, show_comments: bool = True) -> str:
        """ The same as calling str(Entry), except that you can pass options
        to customize how the entry is printed.

        skip_empty_loops will omit printing loops with no tags at all. (A loop with null tags is not "empty".)
        skip_empty_tags will omit tags in the saveframes and loops which have no non-null values.
        show_comments will show the standard comments before a saveframe."""

        return self.__str__(skip_empty_loops=skip_empty_loops, skip_empty_tags=skip_empty_tags,
                            show_comments=show_comments)

    def get_json(self, serialize: bool = True) -> Union[dict, str]:
        """ Returns the entry in JSON format. If serialize is set to
        False a dictionary representation of the entry that is
        serializeable is returned instead."""

        frames = [x.get_json(serialize=False) for x in self.frame_list]

        entry_dict = {
            "entry_id": self.entry_id,
            "saveframes": frames
        }

        if serialize:
            return json.dumps(entry_dict, default=_json_serialize)
        else:
            return entry_dict

    def get_loops_by_category(self, value: str) -> List['loop_mod.Loop']:
        """Allows fetching loops by category."""

        value = utils.format_category(value).lower()

        results = []
        for frame in self.frame_list:
            for one_loop in frame.loops:
                if one_loop.category.lower() == value:
                    results.append(one_loop)
        return results

    def get_saveframe_by_name(self, saveframe_name: str) -> 'saveframe_mod.Saveframe':
        """Allows fetching a saveframe by name."""

        frames = self.frame_dict
        if saveframe_name in frames:
            return frames[saveframe_name]
        else:
            raise KeyError("No saveframe with name '%s'" % saveframe_name)

    def get_saveframes_by_category(self, value: str) -> List['saveframe_mod.Saveframe']:
        """Allows fetching saveframes by category."""

        return self.get_saveframes_by_tag_and_value("sf_category", value)

    def get_saveframes_by_tag_and_value(self, tag_name: str, value: Any) -> List['saveframe_mod.Saveframe']:
        """Allows fetching saveframe(s) by tag and tag value."""

        ret_frames = []

        for frame in self.frame_list:
            results = frame.get_tag(tag_name)
            if results != [] and results[0] == value:
                ret_frames.append(frame)

        return ret_frames

    def get_tag(self, tag: str, whole_tag: bool = False) -> list:
        """ Given a tag (E.g. _Assigned_chem_shift_list.Data_file_name)
        return a list of all values for that tag. Specify whole_tag=True
        and the [tag_name, tag_value (,tag_line_number)] pair will be
        returned."""

        if "." not in str(tag):
            raise ValueError("You must provide the tag category to call this method at the entry level.")

        results = []
        for frame in self.frame_list:
            results.extend(frame.get_tag(tag, whole_tag=whole_tag))

        return results

    def get_tags(self, tags: list) -> Dict[str, list]:
        """ Given a list of tags, get all of the tags and return the
        results in a dictionary."""

        # All tags
        if tags is None or not isinstance(tags, list):
            raise ValueError("Please provide a list of tags.")

        results = {}
        for tag in tags:
            results[tag] = self.get_tag(tag)

        return results

    def normalize(self, schema: Optional['Schema'] = None) -> None:
        """ Sorts saveframes, loops, and tags according to the schema
        provided (or BMRB default if none provided).

        Also re-assigns ID tag values and updates tag links to ID values."""

        # Assign all the ID tags, and update all links to ID tags
        my_schema = utils.get_schema(schema)

        # Sort the saveframes according to ID, if an ID exists. Otherwise, still sort by category
        ordering = my_schema.category_order

        def sf_key(_: saveframe_mod.Saveframe) -> [int, Union[int, float]]:
            """ Helper function to sort the saveframes.
            Returns (category order, saveframe order) """

            # If not a real category, generate an artificial but stable order > the real saveframes
            try:
                category_order = ordering.index(_.tag_prefix)
            except (ValueError, KeyError):
                if _.category is None:
                    category_order = float('infinity')
                else:
                    category_order = len(ordering) + abs(int(hashlib.sha1(str(_.category).encode()).hexdigest(), 16))

            # See if there is an ID tag, and it is a number
            saveframe_id = float('infinity')
            try:
                saveframe_id = int(_.get_tag("ID")[0])
            except (ValueError, KeyError, IndexError, TypeError):
                # Either there is no ID, or it is not a number. By default it will sort at the end of saveframes of its
                # category. Note that the entry_information ID tag has a different meaning, but since there should
                # only ever be one saveframe of that category, the sort order for it can be any value.
                pass

            return category_order, saveframe_id

        def loop_key(_) -> Union[int, float]:
            """ Helper function to sort the loops."""

            try:
                return ordering.index(_.category)
            except ValueError:
                # Generate an arbitrary sort order for loops that aren't in the schema but make sure that they
                #  always come after loops in the schema
                return len(ordering) + abs(int(hashlib.sha1(str(_.category).encode()).hexdigest(), 16))

        # Go through all the saveframes
        for each_frame in self.frame_list:
            each_frame.sort_tags(schema=my_schema)
            # Iterate through the loops
            for each_loop in each_frame:
                each_loop.sort_tags(schema=my_schema)

                # See if we can sort the rows (in addition to tags)
                try:
                    each_loop.sort_rows("Ordinal")
                except ValueError:
                    pass
            each_frame.loops.sort(key=loop_key)
        self.frame_list.sort(key=sf_key)

        # Calculate all the categories present
        categories: set = set()
        for each_frame in self.frame_list:
            categories.add(each_frame.category)

        # tag_prefix -> tag -> original value -> mapped value
        mapping: dict = {}

        # Reassign the ID tags first
        for each_category in categories:

            # First in the saveframe tags
            id_counter: int = 1
            for each_frame in self.get_saveframes_by_category(each_category):
                for tag in each_frame.tags:
                    tag_schema = my_schema.schema.get(f"{each_frame.tag_prefix}.{tag[0]}".lower())
                    if not tag_schema:
                        continue

                    # Make sure the capitalization of the tag is correct
                    tag[0] = tag_schema['Tag field']

                    if tag_schema['lclSfIdFlg'] == 'Y':
                        # If it's an Entry_ID tag, set it that way
                        if tag_schema['entryIdFlg'] == 'Y':
                            mapping[f'{each_frame.tag_prefix[1:]}.{tag[0]}.{tag[1]}'] = self._entry_id
                            tag[1] = self._entry_id
                        # Must be an integer to avoid renumbering the chem_comp ID, for example
                        elif tag_schema['BMRB data type'] == "int":
                            mapping[f'{each_frame.tag_prefix[1:]}.{tag[0]}.{tag[1]}'] = id_counter
                            tag[1] = id_counter
                        # We need to still store all the other tag values too
                        else:
                            mapping[f'{each_frame.tag_prefix[1:]}.{tag[0]}.{tag[1]}'] = tag[1]
                    else:
                        mapping[f'{each_frame.tag_prefix[1:]}.{tag[0]}.{tag[1]}'] = tag[1]

                # Then in the loop
                for loop in each_frame:
                    for x, tag in enumerate(loop.tags):
                        tag_schema = my_schema.schema.get(f"{loop.category}.{tag}".lower())
                        if not tag_schema:
                            continue

                        # Make sure the tags have the proper capitalization
                        loop.tags[x] = tag_schema['Tag field']

                        for row in loop.data:
                            # We don't re-map loop IDs, but we should still store them
                            mapping[f'{loop.category[1:]}.{tag}.{row[x]}'] = row[x]

                            if tag_schema['lclSfIdFlg'] == 'Y':
                                # If it's an Entry_ID tag, set it that way
                                if tag_schema['entryIdFlg'] == 'Y':
                                    row[x] = self.entry_id
                                # Must be an integer to avoid renumbering the chem_comp ID, for example
                                elif tag_schema['BMRB data type'] == "int":
                                    if row[x] in definitions.NULL_VALUES:
                                        row[x] = id_counter
                                # Handle chem_comp and it's ilk
                                else:
                                    parent_id_tag = f"{tag_schema['Foreign Table']}.{tag_schema['Foreign Column']}"
                                    parent_id_value = each_frame[parent_id_tag][0]
                                    row[x] = parent_id_value
                id_counter += 1

        # Now fix any other references
        for saveframe in self:
            for tag in saveframe.tags:
                tag_schema = my_schema.schema.get(f"{saveframe.tag_prefix}.{tag[0]}".lower())
                if not tag_schema:
                    continue
                if tag_schema['Foreign Table'] and tag_schema['Sf pointer'] != 'Y':

                    if tag[1] in definitions.NULL_VALUES:
                        if tag_schema['Nullable']:
                            continue
                        else:
                            logging.warning("A foreign key tag that is not nullable was set to "
                                            f"a null value. Tag: {saveframe.tag_prefix}.{tag[1]} Primary key: "
                                            f"{tag_schema['Foreign Table']}.{tag_schema['Foreign Column']} "
                                            f"Value: {tag[1]}")

                    try:
                        tag[1] = mapping[f"{tag_schema['Foreign Table']}.{tag_schema['Foreign Column']}.{tag[1]}"]
                    except KeyError:
                        logging.warning(f'The tag {saveframe.tag_prefix}.{tag[0]} has value {tag[1]} '
                                        f'but there is no valid primary key.')

            # Now apply the remapping to loops...
            for loop in saveframe:
                for x, tag in enumerate(loop.tags):
                    tag_schema = my_schema.schema.get(f"{loop.category}.{tag}".lower())
                    if not tag_schema:
                        continue
                    if tag_schema['Foreign Table'] and tag_schema['Sf pointer'] != 'Y':
                        for row in loop.data:
                            if row[x] in definitions.NULL_VALUES:
                                if tag_schema['Nullable']:
                                    continue
                                else:
                                    logging.warning("A foreign key reference tag that is not nullable was set to "
                                                    f"a null value. Tag: {loop.category}.{tag} Foreign key: "
                                                    f"{tag_schema['Foreign Table']}.{tag_schema['Foreign Column']} "
                                                    f"Value: {row[x]}")
                            try:
                                row[x] = mapping[
                                    f"{tag_schema['Foreign Table']}.{tag_schema['Foreign Column']}.{row[x]}"]
                            except KeyError:
                                if (loop.category == '_Atom_chem_shift' or loop.category == '_Entity_comp_index') and \
                                        (tag == 'Atom_ID' or tag == 'Comp_ID'):
                                    continue
                                logging.warning(f'The tag {loop.category}.{tag} has value {row[x]} '
                                                f'but there is no valid primary key '
                                                f"{tag_schema['Foreign Table']}.{tag_schema['Foreign Column']} "
                                                f"with the tag value.")

                    # If there is both a label tag and an ID tag, do the reassignment

                    # We found a framecode reference
                    if tag_schema['Foreign Table'] and tag_schema['Foreign Column'] == 'Sf_framecode':

                        # Check if there is a tag pointing to the 'ID' tag
                        for conditional_tag in loop.tags:
                            conditional_tag_schema = my_schema.schema.get(f"{loop.category}.{conditional_tag}".lower())
                            if not conditional_tag_schema:
                                continue
                            if conditional_tag_schema['Foreign Table'] == tag_schema['Foreign Table'] and \
                                    conditional_tag_schema['Foreign Column'] == 'ID' and \
                                    conditional_tag_schema['entryIdFlg'] != 'Y':
                                # We found the matching tag
                                tag_pos = loop.tag_index(conditional_tag)

                                for row in loop.data:
                                    # Check if the tag is null
                                    if row[x] in definitions.NULL_VALUES:
                                        if tag_schema['Nullable']:
                                            continue
                                        else:
                                            logging.info(f"A foreign saveframe reference tag that is not nullable was "
                                                         f"set to a null value. Tag: {loop.category}.{tag} "
                                                         "Foreign saveframe: "
                                                         f"{tag_schema['Foreign Table']}.{tag_schema['Foreign Column']}"
                                                         )
                                            continue
                                    try:
                                        row[tag_pos] = self.get_saveframe_by_name(row[x][1:])['ID'][0]
                                    except IndexError:
                                        logging.info(f"Getting {self.get_saveframe_by_name(row[x][1:])['ID']}")
                                    except KeyError:
                                        logging.warning(f"Missing frame of type {tag} pointed to by {conditional_tag}")

        # Renumber the 'ID' column in a loop
        for each_frame in self.frame_list:
            for loop in each_frame.loops:
                if loop.tag_index('ID') is not None and loop.category != '_Experiment':
                    loop.renumber_rows('ID')

    def print_tree(self) -> None:
        """Prints a summary, tree style, of the frames and loops in
        the entry."""

        print(repr(self))
        frame: saveframe_mod.Saveframe
        for pos, frame in enumerate(self):
            print("\t[%d] %s" % (pos, repr(frame)))
            for pos2, one_loop in enumerate(frame):
                print("\t\t[%d] %s" % (pos2, repr(one_loop)))

    def rename_saveframe(self, original_name: str, new_name: str) -> None:
        """ Renames a saveframe and updates all pointers to that
        saveframe in the entry with the new name."""

        # Strip off the starting $ in the names
        if original_name.startswith("$"):
            original_name = original_name[1:]
        if new_name.startswith("$"):
            new_name = new_name[1:]

        # Make sure there is no saveframe called what the new name is
        if [x.name for x in self.frame_list].count(new_name) > 0:
            raise ValueError("Cannot rename a saveframe as '%s' because a saveframe with that name already exists." %
                             new_name)

        # This can raise a ValueError, but no point catching it since it really is a ValueError if they provide a name
        #  of a saveframe that doesn't exist in the entry.
        change_frame = self.get_saveframe_by_name(original_name)

        # Update the saveframe
        change_frame.name = new_name

        # What the new references should look like
        old_reference = "$" + original_name
        new_reference = "$" + new_name

        # Go through all the saveframes
        for each_frame in self:
            # Iterate through the tags
            for each_tag in each_frame.tags:
                if each_tag[1] == old_reference:
                    each_tag[1] = new_reference
            # Iterate through the loops
            for each_loop in each_frame:
                for each_row in each_loop:
                    for pos, val in enumerate(each_row):
                        if val == old_reference:
                            each_row[pos] = new_reference

    def validate(self, validate_schema: bool = True, schema: 'Schema' = None,
                 validate_star: bool = True) -> List[str]:
        """Validate an entry in a variety of ways. Returns a list of
        errors found. 0-length list indicates no errors found. By
        default all validation modes are enabled.

        validate_schema - Determines if the entry is validated against
        the NMR-STAR schema. You can pass your own custom schema if desired,
        otherwise the cached schema will be used.

        validate_star - Determines if the STAR syntax checks are ran."""

        errors = []

        # They should validate for something...
        if not validate_star and not validate_schema:
            errors.append("Validate() should be called with at least one validation method enabled.")

        if validate_star:

            # Check for saveframes with same name
            saveframe_names = sorted(x.name for x in self)
            for ordinal in range(0, len(saveframe_names) - 2):
                if saveframe_names[ordinal] == saveframe_names[ordinal + 1]:
                    errors.append("Multiple saveframes with same name: '%s'" % saveframe_names[ordinal])

            # Check for dangling references
            fdict = self.frame_dict

            for each_frame in self:
                # Iterate through the tags
                for each_tag in each_frame.tags:
                    tag_copy = str(each_tag[1])
                    if (tag_copy.startswith("$")
                            and tag_copy[1:] not in fdict):
                        errors.append("Dangling saveframe reference '%s' in tag '%s.%s'" % (each_tag[1],
                                                                                            each_frame.tag_prefix,
                                                                                            each_tag[0]))

                # Iterate through the loops
                for each_loop in each_frame:
                    for each_row in each_loop:
                        for pos, val in enumerate(each_row):
                            val = str(val)
                            if val.startswith("$") and val[1:] not in fdict:
                                errors.append("Dangling saveframe reference '%s' in tag '%s.%s'" %
                                              (val,
                                               each_loop.category,
                                               each_loop.tags[pos]))

        # Ask the saveframes to check themselves for errors
        for frame in self:
            errors.extend(frame.validate(validate_schema=validate_schema, schema=schema, validate_star=validate_star))

        return errors

    def write_to_file(self, file_name: str, format_: str = "nmrstar", show_comments: bool = True,
                      skip_empty_loops: bool = False, skip_empty_tags: bool = False) -> None:
        """ Writes the entry to the specified file in NMR-STAR format.

        Optionally specify:
        show_comments=False to disable the comments that are by default inserted. Ignored when writing json.
        skip_empty_loops=False to force printing loops with no tags at all (loops with null tags are still printed)
        skip_empty_tags=True will omit tags in the saveframes and loops which have no non-null values.
        format_=json to write to the file in JSON format."""

        if format_ not in ["nmrstar", "json"]:
            raise ValueError("Invalid output format.")

        data_to_write = ''
        if format_ == "nmrstar":
            data_to_write = self.format(show_comments=show_comments, skip_empty_loops=skip_empty_loops,
                                        skip_empty_tags=skip_empty_tags)
        elif format_ == "json":
            data_to_write = self.get_json()

        out_file = open(file_name, "w")
        out_file.write(data_to_write)
        out_file.close()



import decimal
import json
import logging
import os
import sys
import time
import zlib
from datetime import date
from gzip import GzipFile
from io import StringIO, BytesIO
from typing import Dict, Union, IO
from urllib.error import HTTPError, URLError
from urllib.request import urlopen, Request

import pynmrstar

__version__: str = "3.1.0"

# If we have requests, open a session to reuse for the duration of the program run
try:
    from requests import session as _requests_session
    # This replaces the urllib HTTPError if we have requests
    from requests.exceptions import HTTPError
    _session = _requests_session()
except ModuleNotFoundError:
    _session = None


def _build_extension() -> bool:
    """ Try to compile the c extension. """
    import subprocess

    cur_dir = os.getcwd()
    try:
        src_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)))
        os.chdir(os.path.join(src_dir, '..', "c"))

        # Use the appropriate build command
        process = subprocess.Popen(['make', 'python3'], stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
        stdout, stderr = process.communicate()
        ret_code = process.poll()
        # The make command exited with a non-zero status
        if ret_code:
            logging.warning('Compiling cnmrstar failed with error code %s and stderr: %s', ret_code, stderr)
            return False

        # We were able to build the extension?
        return True
    except OSError:
        # There was an error going into the c dir
        logging.warning('Could not find a directory with c source code.')
        return False
    finally:
        # Go back to the directory we were in before exiting
        os.chdir(cur_dir)


def _get_cnmrstar() -> Union[None, object]:
    """ Returns the cnmrstar module, or returns None if it isn't available. """

    # First see if it's installed via pip
    try:
        import cnmrstar
        logging.debug('Imported cnmrstar via installed package.')
        return cnmrstar
    except ImportError:

        # See if it is compiled locally
        try:
            import pynmrstar.cnmrstar as cnmrstar
            logging.debug('Imported cnmrstar from locally compiled file.')

            if "version" not in dir(cnmrstar) or cnmrstar.version() < "3.0.9":
                logging.warning("Recompiling cnmrstar module due to API changes. You may experience a segmentation "
                                "fault immediately following this message but should have no issues the next time you "
                                "run your script or this program.")
                _build_extension()
                sys.exit(0)

            return cnmrstar

        except ImportError:

            # Try to compile cnmrstar, but check for the 'no c module' file before continuing
            if not os.path.isfile(os.path.join(os.path.dirname(os.path.realpath(__file__)), ".nocompile")):
                logging.info('Compiling local cnmrstar module...')

                if _build_extension():
                    try:
                        import pynmrstar.cnmrstar as cnmrstar
                        logging.debug('Imported cnmrstar from locally compiled file.')
                        return cnmrstar
                    except ImportError:
                        return
            else:
                logging.debug("Not compiling cnmrstar due to presence of '.nocompile' file")
                return


# noinspection PyDefaultArgument
def _get_comments(_comment_cache: Dict[str, Dict[str, str]] = {}) -> Dict[str, Dict[str, str]]:
    """ Loads the comments that should be placed in written files.

    The default argument is mutable on purpose, as it is used as a cache for memoization."""

    # Comment dictionary already exists
    if _comment_cache:
        return _comment_cache

    file_to_load = os.path.join(os.path.dirname(os.path.realpath(__file__)))
    file_to_load = os.path.join(file_to_load, "reference_files/comments.str")

    # The import needs to be here to avoid import errors due to circular imports
    from pynmrstar.entry import Entry
    try:
        comment_entry = Entry.from_file(file_to_load)
    except IOError:
        # Load the comments from Github if we can't find them locally
        try:
            logging.warning('Could not load comments from disk. Loading from web...')
            comment_entry = Entry.from_file(_interpret_file(pynmrstar.definitions.COMMENT_URL))
        except Exception:
            logging.exception('Could not load comments from web. No comments will be shown.')
            # No comments will be printed
            return {}

    # Load the comments
    comment_records = comment_entry[0][0].get_tag(["category", "comment", "every_flag"])
    comment_map = {'N': False, 'Y': True}
    for comment in comment_records:
        if comment[1] != ".":
            _comment_cache[comment[0]] = {'comment': comment[1].rstrip() + "\n\n",
                                          'every_flag': comment_map[comment[2]]}

    return _comment_cache


def _json_serialize(obj: object) -> str:
    """JSON serializer for objects not serializable by default json code"""

    # Serialize datetime.date objects by calling str() on them
    if isinstance(obj, (date, decimal.Decimal)):
        return str(obj)
    raise TypeError("Type not serializable: %s" % type(obj))


def _get_url_reliably(url: str, wait_time: int = 5, raw: bool = False, timeout: int = 10):
    """ Attempts to load data from a URL, retrying two times with an exponential
    backoff if rate limited. Fails immediately on 4xx errors that are not 403."""

    # If using Requests
    if _session:
        response = _session.get(url, timeout=timeout,
                                headers={'Application': f'PyNMRSTAR {__version__}'})
        # We are rate limited - sleep and try again
        if response.status_code == 403:
            if wait_time <= 25:
                logging.warning(f'We were rate limited. Sleeping for {wait_time} seconds.')
                time.sleep(wait_time)
                return _get_url_reliably(url, wait_time*5)
            else:
                raise HTTPError("Continued to receive 403 after multiple wait times.")
        if response.status_code == 404:
            raise KeyError(f"Server returned 404.")
        response.raise_for_status()
        if raw:
            return response.content
        else:
            return response.text
    else:
        # Use the built in library
        try:
            req = Request(url)
            req.add_header('Application', f'PyNMRSTAR {__version__}')
            url_request = urlopen(req, timeout=timeout)
            serialized_ent = url_request.read()
            url_request.close()

        except HTTPError as err:
            if err.code == 404:
                raise KeyError(f"Server returned 404.")
            # We are rate limited - sleep and try again
            elif err.code == 403:
                if wait_time <= 25:
                    logging.warning(f'We were rate limited. Sleeping for {wait_time} seconds.')
                    time.sleep(wait_time)
                    return _get_url_reliably(url, wait_time * 5)
                else:
                    raise err
            else:
                raise err
        if raw:
            return serialized_ent
        else:
            return serialized_ent.decode()


def _get_entry_from_database(entry_num: Union[str, int], convert_data_types: bool = False) -> 'pynmrstar.Entry':
    """ Fetches an entry from the API (or falls back to the FTP site) in
    as reliable and robust a way as possible. Used by Entry.from_database(). """

    entry_num = str(entry_num).lower()
    if entry_num.startswith("bmr"):
        entry_num = entry_num[3:]

    # Try to load the entry using JSON

    entry_url: str = (pynmrstar.definitions.API_URL + "/entry/%s?format=zlib") % entry_num

    try:
        serialized_ent = _get_url_reliably(entry_url, raw=True)
        json_data = json.loads(zlib.decompress(serialized_ent).decode())
        if "error" in json_data:
            raise RuntimeError('Something wrong with API response.')
        ent = pynmrstar.Entry.from_json(json_data)
    except (HTTPError, HTTPError, RuntimeError):
        # Can't fall back to FTP for chemcomps
        if entry_num.startswith("chemcomp"):
            raise IOError("Unable to load that chemcomp from the API.")

        # We're going to try again from the FTP
        logging.warning('Failed to download entry from the API, trying again from the FTP site.')
        if "bmse" in entry_num or "bmst" in entry_num:
            url = f"{pynmrstar.definitions.FTP_URL}/metabolomics/entry_directories/{entry_num}/{entry_num}.str"
        else:
            url = f"{pynmrstar.definitions.FTP_URL}/entry_directories/bmr{entry_num}/bmr{entry_num}_3.str"
        try:
            # Use a longer timeout for the timeout
            entry_content = _get_url_reliably(url, raw=False, timeout=20)
            ent = pynmrstar.Entry.from_string(entry_content)
        except (HTTPError, HTTPError):
            raise IOError(f"Entry {entry_num} does not exist in the public database.")
        except URLError:
            raise IOError("You don't appear to have an active internet connection. Cannot fetch entry.")

    except KeyError:
        raise IOError(f"Entry {entry_num} does not exist in the public database.")

    # Update the entry source
    ent.source = f"from_database({entry_num})"
    for each_saveframe in ent:
        each_saveframe.source = ent.source
        for each_loop in each_saveframe:
            each_loop.source = ent.source

    if convert_data_types:
        schema = pynmrstar.utils.get_schema()
        for each_saveframe in ent:
            for tag in each_saveframe.tags:
                cur_tag = each_saveframe.tag_prefix + "." + tag[0]
                tag[1] = schema.convert_tag(cur_tag, tag[1], line_num="SF %s" % each_saveframe.name)
            for loop in each_saveframe:
                for row in loop.data:
                    for pos in range(0, len(row)):
                        category = loop.category + "." + loop.tags[pos]
                        line_num = "Loop %s" % loop.category
                        row[pos] = schema.convert_tag(category, row[pos], line_num=line_num)

    return ent


def _interpret_file(the_file: Union[str, IO]) -> StringIO:
    """Helper method returns some sort of object with a read() method.
    the_file could be a URL, a file location, a file object, or a
    gzipped version of any of the above."""

    if hasattr(the_file, 'read'):
        read_data: Union[bytes, str] = the_file.read()
        if type(read_data) == bytes:
            buffer: BytesIO = BytesIO(read_data)
        elif type(read_data) == str:
            buffer = BytesIO(read_data.encode())
        else:
            raise IOError("What did your file object return when .read() was called on it?")
    elif isinstance(the_file, str):
        if the_file.startswith("http://") or the_file.startswith("https://") or the_file.startswith("ftp://"):
            buffer = BytesIO(_get_url_reliably(the_file, raw=True))
        else:
            with open(the_file, 'rb') as read_file:
                buffer = BytesIO(read_file.read())
    else:
        raise ValueError("Cannot figure out how to interpret the file you passed.")

    # Decompress the buffer if we are looking at a gzipped file
    try:
        gzip_buffer = GzipFile(fileobj=buffer)
        gzip_buffer.readline()
        gzip_buffer.seek(0)
        buffer = BytesIO(gzip_buffer.read())
    # Apparently we are not looking at a gzipped file
    except (IOError, AttributeError, UnicodeDecodeError):
        pass

    buffer.seek(0)
    return StringIO(buffer.read().decode())



#!/usr/bin/env python3

#############################################
#                 Imports                   #
#############################################

import json
import os
from typing import Iterable, Any, Dict
from urllib.error import HTTPError, URLError

from pynmrstar import definitions, entry as entry_mod
from pynmrstar._internal import _interpret_file
from pynmrstar.schema import Schema

try:
    import cnmrstar
except ImportError:
    cnmrstar = None

#############################################
#            Global Variables               #
#############################################

# Set this to allow import * from pynmrstar to work sensibly
__all__ = ['diff', 'format_category', 'format_tag', 'get_schema', 'iter_entries', 'quote_value', 'validate']


def diff(entry1: 'entry_mod.Entry', entry2: 'entry_mod.Entry') -> None:
    """Prints the differences between two entries. Non-equal entries
    will always be detected, but specific differences detected depends
    on order of entries."""

    diffs = entry1.compare(entry2)
    if len(diffs) == 0:
        print("Identical entries.")
    for difference in diffs:
        print(difference)


def format_category(tag: str) -> str:
    """Adds a '_' to the front of a tag (if not present) and strips out
    anything after a '.'"""

    if tag:
        if not tag.startswith("_"):
            tag = "_" + tag
        if "." in tag:
            tag = tag[:tag.index(".")]
    return tag


def format_tag(tag: str) -> str:
    """Strips anything before the '.'"""

    if '.' in tag:
        return tag[tag.index('.') + 1:]
    return tag


# noinspection PyDefaultArgument
def get_schema(passed_schema: 'Schema' = None, _cached_schema: Dict[str, Schema] = {}) \
        -> 'Schema':
    """If passed a schema (not None) it returns it. If passed none,
    it checks if the default schema has been initialized. If not
    initialized, it initializes it. Then it returns the default schema."""

    if passed_schema:
        return passed_schema

    if not _cached_schema:

        # Try to load the local file first
        try:
            schema_file = os.path.join(os.path.dirname(os.path.realpath(__file__)))
            schema_file = os.path.join(schema_file, "reference_files/schema.csv")
            _cached_schema['schema'] = Schema(schema_file=schema_file)
        except IOError:
            # Try to load from the internet
            try:
                _cached_schema['schema'] = Schema()
            except (HTTPError, URLError):
                raise ValueError("Could not load a BMRB schema from the internet or from the local repository.")

    return _cached_schema['schema']


def iter_entries(metabolomics: bool = False) -> Iterable['entry_mod.Entry']:
    """ Returns a generator that will yield an Entry object for every
        macromolecule entry in the current BMRB database. Perfect for performing
        an operation across the entire BMRB database. Set `metabolomics=True`
        in order to get all the entries in the metabolomics database."""

    api_url = "%s/list_entries?database=macromolecules" % definitions.API_URL
    if metabolomics:
        api_url = "%s/list_entries?database=metabolomics" % definitions.API_URL

    for entry in json.loads(_interpret_file(api_url).read()):
        yield entry_mod.Entry.from_database(entry)


def quote_value(value: Any) -> str:
    """Automatically quotes the value in the appropriate way. Don't
    quote values you send to this method or they will show up in
    another set of quotes as part of the actual data. E.g.:

    clean_value('"e. coli"') returns '\'"e. coli"\''

    while

    clean_value("e. coli") returns "'e. coli'"

    This will automatically be called on all values when you use a str()
    method (so don't call it before inserting values into tags or loops).

    Be mindful of the value of STR_CONVERSION_DICT as it will effect the
    way the value is converted to a string.

    """

    # Allow manual specification of conversions for booleans, Nones, etc.
    if value in definitions.STR_CONVERSION_DICT:
        if any(isinstance(value, type(x)) for x in definitions.STR_CONVERSION_DICT):
            value = definitions.STR_CONVERSION_DICT[value]

    # Use the fast code if it is available
    if cnmrstar is not None:
        # It's faster to assume we are working with a string and catch
        #  errors than to check the instance for every object and convert
        try:
            return cnmrstar.clean_value(value)
        except (ValueError, TypeError):
            return cnmrstar.clean_value(str(value))

    # Convert non-string types to string
    if not isinstance(value, str):
        value = str(value)

    # If it is a STAR-format multi-line comment already, we need to escape it
    if "\n;" in value:
        value = value.replace("\n", "\n   ")
        if value[-1] != "\n":
            value = value + "\n"
        if value[0] != "\n":
            value = "\n   " + value
        return value

    # If it's going on it's own line, don't touch it
    if "\n" in value:
        if value[-1] != "\n":
            return value + "\n"
        return value

    if value == "":
        raise ValueError("Empty strings are not allowed as values. Use a '.' or a '?' if needed.")

    # If it has single and double quotes it will need to go on its
    #  own line under certain conditions...
    if '"' in value and "'" in value:
        can_wrap_single = True
        can_wrap_double = True

        for pos, char in enumerate(value):
            next_char = value[pos + 1:pos + 2]

            if next_char != "" and next_char in definitions.WHITESPACE:
                if char == "'":
                    can_wrap_single = False
                if char == '"':
                    can_wrap_double = False

        if not can_wrap_single and not can_wrap_double:
            return '%s\n' % value
        elif can_wrap_single:
            return "'%s'" % value
        elif can_wrap_double:
            return '"%s"' % value

    # Check for special characters in a tag
    if any(x in value for x in definitions.WHITESPACE) or '#' in value or value in definitions.RESERVED_KEYWORDS or \
            value.startswith("_"):
        # If there is a single quote wrap in double quotes
        if "'" in value:
            return '"%s"' % value
        # Either there is a double quote or no quotes
        else:
            return "'%s'" % value

    # Quote if necessary
    if value[0] == "'":
        return '"' + value + '"'
    if value[0] == '"':
        return "'" + value + "'"

    # It's good to go
    return value


def validate(entry_to_validate: 'entry_mod.Entry', schema: 'Schema' = None) -> None:
    """Prints a validation report of an object."""

    validation = entry_to_validate.validate(schema=schema)
    if len(validation) == 0:
        print("No problems found during validation.")
    for pos, err in enumerate(validation):
        print("%d: %s" % (pos + 1, err))
